"""finders are used by Geometry.find to locate atoms in a more general way"""
import numpy as np


class Finder:
    def get_matching_atoms(self, atoms, geometry=None):
        """overwrite with function that returns list(Atom) of the atoms that
        match your Finder's criteria
        geometry is an optional argument that could be used to e.g. find
        atoms a certain number of bonds"""
        pass


class BondsFrom(Finder):
    """exact number of bonds from specified atom
    avoid: bonding path cannot pass through these atoms"""
    def __init__(self, atom, number_of_bonds, avoid=None):
        super().__init__()

        self.central_atom = atom
        self.n_bonds = number_of_bonds
        self.avoid = avoid

    def __repr__(self):
        return "atoms %i bonds of %s" % (self.n_bonds, self.central_atom)

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that are a certain number of bonds away from the given atom"""
        matching_atoms = []
        for atom in atoms:
            try:
                path = geometry.shortest_path(atom, self.central_atom, avoid=self.avoid)
            except LookupError:
                continue

            if len(path) - 1 == self.n_bonds:
                matching_atoms.append(atom)

        return matching_atoms


class WithinBondsOf(BondsFrom):
    """within a specified number of bonds from the atom"""
    def __init__(self, atom, number_of_bonds):
        super().__init__(atom, number_of_bonds)

    def __repr__(self):
        return "atoms %i bonds of %s" % (self.n_bonds, self.central_atom)

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that are a certain number of bonds away from the given atom"""
        matching_atoms = []
        for atom in atoms:
            try:
                path = geometry.shortest_path(atom, self.central_atom)
            except LookupError:
                continue

            if len(path) - 1 <= self.n_bonds and len(path) > 1:
                matching_atoms.append(atom)

        return matching_atoms


class BondedTo(Finder):
    """returns all atoms that are bonded to the specified atom"""
    def __init__(self, atom):
        super().__init__()

        self.atom = atom

    def __repr__(self):
        return "atoms bonded to %s" % self.atom

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        return [atom for atom in atoms if atom in self.atom.connected]


class WithinRadiusFromPoint(Finder):
    """within a specified radius of a point"""
    def __init__(self, point, radius):
        super().__init__()

        self.point = np.array(point)
        self.radius = radius
    
    def __repr__(self):
        return "atoms within %.2f angstroms of (%.2f, %.2f, %.2f)" % (self.radius, *self.point)

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        matching_atoms = []
        for atom in atoms:
            coords = atom.coords
            d = np.linalg.norm(coords - self.point)
            if d < self.radius:
                matching_atoms.append(atom)

        return matching_atoms


class WithinRadiusFromAtom(Finder):
    """within a specified radius of a point"""
    def __init__(self, atom, radius):
        super().__init__()

        self.atom = atom
        self.radius = radius
    
    def __repr__(self):
        return "atoms within %.2f angstroms of %s" % (self.radius, self.atom)

    def get_matching_atoms(self, atoms, geometry=None):
        """returns list(Atom) that are within a radius of a point"""
        matching_atoms = []
        for atom in atoms:
            coords = atom.coords
            d = self.atom.dist(atom)
            if d < self.radius:
                matching_atoms.append(atom)

        return matching_atoms
    

class NotAny(Finder):
    """atoms not matching specifiers/Finders"""
    def __init__(self, *args):
        """args can be any number of Finders and/or other atom specifiers (tags, elements, etc.)"""
        super().__init__()

        self.critera = args

    def __repr__(self):
        return "not any of: %s" % ", ".join([str(x) for x in self.critera])

    def get_matching_atoms(self, atoms, geometry):
        """returns List(Atom) that do not match any of the critera"""
        unmatching_atoms = []
        for criterion in self.critera:
            try:
                unmatch = geometry.find(criterion)
                unmatching_atoms.extend(unmatch)
            
            except LookupError:
                pass

        return [atom for atom in atoms if atom not in set(unmatching_atoms)]


class AnyTransitionMetal(Finder):
    """any atoms that are transition metals"""
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "any transition metal"

    def get_matching_atoms(self, atoms, geometry=None):
        """returns List(Atom) of atoms that are metals"""
        from AaronTools.const import TMETAL
        return [atom for atom in atoms if atom.element in TMETAL]


class AnyNonTransitionMetal(NotAny):
    """any atoms that are not transition metals"""
    def __init__(self):
        super().__init__(AnyTransitionMetal())

    def __repr__(self):
        return "any non-transition metal"


class HasAttribute(Finder):
    """all atoms with the specified attribute"""
    def __init__(self, attribute):
        super().__init__()

        self.attribute_name = attribute

    def __repr__(self):
        return "atoms with the '%s' attribute" % self.attribute_name

    def get_matching_atoms(self, atoms, geometry=None):
        """returns List(Atom) of atoms that have the attribute"""
        return [atom for atom in atoms if hasattr(atom, self.attribute_name)]


class VSEPR(Finder):
    """atoms with the specified VSEPR geometry
    see Atom.get_shape for a list of valid vsepr_geometry strings"""
    def __init__(self, vsepr_geometry):
        super().__init__()
        
        self.vsepr = vsepr_geometry
    
    def __repr__(self):
        return "atoms with %s shape" % self.vsepr

    def get_matching_atoms(self, atoms, geometry=None):
        matching_atoms = []
        for atom in atoms:
            out = atom.get_vsepr()
            if out is not None:
                shape, score = atom.get_vsepr()
                if shape == self.vsepr and score < 0.5:
                    matching_atoms.append(atom)
        
        return matching_atoms


class BondedElements(Finder):
    """atoms bonded to the specified neighboring elements
    if match_exact=True (default), elements must match exactly 
    e.g. BondedElements('C') will find
    atoms bonded to only one carbon and nothing else"""
    def __init__(self, *args, match_exact=True):
        super().__init__()
        
        self.elements = list(args)
        self.match_exact = match_exact
        
    def __repr__(self):
        if len(self.elements) == 0:
            return "atoms bonded to nothing"
        elif len(self.elements) == 1:
            return "atoms bonded to %s" % self.elements[0]
        else:
            return "atoms bonded to %s and %s" % (", ".join(self.elements[:-1]), self.elements[-1])
    
    def get_matching_atoms(self, atoms, geometry=None):
        matching_atoms = []
        if self.match_exact:
            ref = "".join(sorted(self.elements))
        else:
            ref = self.elements
        
        for atom in atoms:
            if self.match_exact:
                ele_list = [a.element for a in [ele for ele in atom.connected]]
                test = "".join(sorted(ele_list))
                if ref == test:
                    matching_atoms.append(atom)
            
            else:
                bonded_eles = [bonded_atom.element for bonded_atom in atom.connected]
                if all([ele in bonded_eles for ele in self.elements]):
                    matching_atoms.append(atom)
        
        return matching_atoms


class NumberOfBonds(Finder):
    """atoms with the specified number of bonds"""
    def __init__(self, num_bonds):
        super().__init__()
        
        self.num_bonds = num_bonds
    
    def __repr__(self):
        return "atoms with %i bonds" % self.num_bonds
    
    def get_matching_atoms(self, atoms, geometry=None):
        return [atom for atom in atoms if len(atom.connected) == self.num_bonds]


class ChiralCentres(Finder):
    """chiral centers
    atoms with a non-planar VSEPR geometry with all bonded groups
    being distinct
    for rings, looks for a set of unique canonical ranks for atoms that 
    are all the same number of bonds away from one atom"""
    #IUPAC spelling 
    def __init__(self, RS_only=False):
        """RS_only: bool  - if True, do not identify chiral centers that are chiral because they
                            are connected to multiple chiral fragments with the same chirality
                            this corresponds to R/S centers
                            False will include r/s centers as well
        """
        super().__init__()
        self.cip = RS_only

    def __repr__(self):
        return "chiral centers"

    def get_matching_atoms(self, atoms, geometry):
        from AaronTools.geometry import Geometry

        matching_atoms = []

        # b/c they are connected to chiral fragments
        geometry.refresh_ranks()
        chiral_atoms_changed = True
        ranks = geometry.canonical_rank(break_ties=False, update=False)
        frags = []
        for atom in geometry.atoms:
            frags.append([])
            for bonded_atom in atom.connected:
                frags[-1].append(geometry.get_fragment(bonded_atom, atom, as_object=False))

        #need to do multiple passes b/c sometimes atoms are chiral
        k = 0
        while chiral_atoms_changed:
            chiral_atoms_changed = False
            k += 1
            #skip atoms we've already found
            for ndx, atom in enumerate(atoms):
                if atom in matching_atoms:
                    continue

                #can't be chiral with 2 bonds
                if len(atom.connected) < 3:
                    continue

                #planar vsepr don't get checked
                vsepr = atom.get_vsepr()
                if vsepr is not None:
                    shape, score = vsepr
                    if shape in  ['trigonal planar', 't shaped', 'sqaure planar']:
                        continue

                chiral = True
                for i, frag1 in enumerate(frags[ndx]):
                    #get the ranks of the atoms in this fragment
                    ranks_1 = [ranks[geometry.atoms.index(atom)] for atom in frag1]
                    for j, frag2 in enumerate(frags[ndx][:i]):
                        same = True
                        
                        ranks_2 = [ranks[geometry.atoms.index(atom)] for atom in frag2]
                        
                        if len(frag1) != len(frag2):
                            same = False
                            continue

                        for a, b in zip(sorted(ranks_1), sorted(ranks_2)):
                            # want correct elements
                            if a != b:
                                same = False
                                break
                           
                        for a, b in zip(sorted(frag1), sorted(frag2)):
                            # and other chiral atoms
                            if not self.cip and a in matching_atoms and b in matching_atoms:
                                #use RMSD to see if they have the same handedness
                                a_connected = sorted(a.connected)
                                b_connected = sorted(b.connected)
                                a_targets = [a] + list(a_connected)
                                b_targets = [b] + list(b_connected)
                                if geometry.RMSD(geometry, targets=a_targets, ref_targets=b_targets, sort=False, align=False) < 0.1:
                                    same = False
                                    break

                            # and correct connected elements
                            for i, j in zip(
                                sorted([aa.element for aa in a.connected]),
                                sorted([bb.element for bb in b.connected]),
                            ):
                                if i != j:
                                    same = False
                                    break

                        
                        ring_atoms = [bonded_atom for bonded_atom in atom.connected if bonded_atom in frag1 and bonded_atom in frag2]
                        if len(ring_atoms) > 0:
                            #this is a ring
                            #look at the rank of all atoms that are n bonds away from this atom
                            #if the ranks are ever all different, this is a chiral center
                            n_bonds = 1
                            acceptable_nbonds = True
                            while acceptable_nbonds:
                                try:
                                    atoms_within_nbonds = geometry.find(BondsFrom(atom, n_bonds))
                                    nbonds_ranks = [ranks[geometry.atoms.index(a)] for a in atoms_within_nbonds]
                                    if all([nbonds_ranks.count(r) == 1 for r in nbonds_ranks]):
                                        same = False
                                        acceptable_nbonds = False
                                    elif not self.cip:
                                        #need to find things in the ring that are chiral b/c of other chiral centers
                                        for i, atom1 in enumerate(atoms_within_nbonds):
                                            for j, atom2 in enumerate(atoms_within_nbonds[i+1:]):
                                                k = j + i + 1
                                                if nbonds_ranks[i] == nbonds_ranks[k]:
                                                    a_connected = sorted(atom1.connected)
                                                    b_connected = sorted(atom2.connected)
                                                    a_targets = [atom1] + list(a_connected)
                                                    b_targets = [atom2] + list(b_connected)
                                                    if geometry.RMSD(geometry, targets=a_targets, ref_targets=b_targets, sort=False, align=False) < 0.1:
                                                        same = False
                                                        break
                                        if not same:
                                            break

                                    n_bonds += 1
                                except LookupError:
                                    acceptable_nbonds = False

                            if not same:
                                break


                        if same:
                            chiral = False
                            break

                if chiral:
                    chiral_atoms_changed = True
                    matching_atoms.append(atom)
        
        return matching_atoms

#alternative spelling
ChiralCenters = ChiralCentres


class FlaggedAtoms(Finder):
    """
    atoms with a non-zero flag
    """
    # useful for finding constrained atoms
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "flagged atoms"

    def get_matching_atoms(self, atoms, geometry):
        return [atom for atom in atoms if atom.flag]


class CloserTo(Finder):
    """
    atoms closer to atom1 than atom2 (based on bonds, not actual distance)
    """
    def __init__(self, atom1, atom2, include_ties=False):
        super().__init__()
        
        self.atom1 = atom1
        self.atom2 = atom2
        self.include_ties = include_ties
    
    def __repr__(self):
        return "atoms closer to %s than %s" % (self.atom1, self.atom2)
    
    def get_matching_atoms(self, atoms, geometry):
        matching_atoms = []
        for atom in atoms:
            if atom is self.atom1 and atom is not self.atom2:
                matching_atoms.append(atom)
                continue
                
            try:
                d1 = len(geometry.shortest_path(self.atom1, atom))
            except LookupError:
                d1 = False
            
            try:
                d2 = len(geometry.shortest_path(self.atom2, atom))
            except LookupError:
                d2 = False
            
            if d1 is not False and d2 is not False and d1 <= d2:
                if self.include_ties:
                    matching_atoms.append(atom)
                elif d1 < d2:
                    matching_atoms.append(atom)
            
            elif d1 is not False and d2 is False:
                matching_atoms.append(atom)
        
        return matching_atoms
    


#!/usr/bin/env python3

import argparse
import os

from sys import stdin
from warnings import warn
from numpy import prod, pi

from AaronTools.fileIO import FileReader, read_types
from AaronTools.finders import BondedTo, ChiralCenters
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent


changechiral_parser = argparse.ArgumentParser(
    description="change handedness of chiral centers",
    formatter_class=argparse.RawTextHelpFormatter
)

changechiral_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[stdin],
    help="a coordinate file"
)

changechiral_parser.add_argument(
    "-i", "--info",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_info",
    help="list information on detected chiral centers"
)

changechiral_parser.add_argument(
    "-if", "--input-format",
    type=str,
    nargs=1,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

changechiral_parser.add_argument(
    "-o", "--output-destination",
    type=str,
    default=None,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$i in the filename will be replaced with a number\n" +
    "if a directory is given, default is \"diastereomer-$i.xyz\" in \n" +
    "that directory\n" +
    "Default: stdout"
)

changechiral_parser.add_argument(
    "-t", "--targets",
    type=str,
    default=None,
    action="append",
    required=False,
    dest="targets",
    help="comma- or hyphen-seperated list of chiral centers to invert (1-indexed)\n" +
    "Chiral centers must have at least two fragments not in a ring\n" +
    "Detected chiral centers are atoms that:\n" +
    "    - have > 2 bonds\n" +
    "    - have a non-planar VSEPR shape\n" +
    "    - each connected fragment is distinct\n" +
    "Default: change chirality of any detected chiral centers"
)

changechiral_parser.add_argument(
    "-d", "--diastereomers",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="combos",
    help="print all diastereomers for selected chiral centers"
)

changechiral_parser.add_argument(
    "-m", "--minimize",
    action="store_true",
    default=False,
    dest="minimize",
    help="rotate substituents to mitigate steric clashing",
)


args = changechiral_parser.parse_args()

s = ""

for infile in args.infile:
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format[0], infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format[0], stdin))
        else:
            f = FileReader(("from stdin", "xyz", stdin))

    geom = Geometry(f)

    target_list = []
    if args.targets is None:
        try:
            target_list = geom.find(ChiralCenters())
        except LookupError as e:
            warn(str(e))
    else:
        for targ in args.targets:
            target_list.extend(geom.find(targ))

    if args.list_info:
        if len(args.infile) > 1:
            s += "%s\n" % infile
        s += "Target\tElement\n"
        for targ in target_list:
            s += "%-2s\t%-2s\n" % (targ.name, targ.element)
        if infile is not args.infile[-1]:
            s += "\n"
        continue

    geom.substituents = []
    if args.combos:
        # this stuff is copy-pasted from makeConf, so it's a bit overkill
        # for getting all diastereomers, as each chiral center can only
        # have 2 options instead of the random number of rotamers
        # a substituent can have
        diastereomers = []
        for targ in target_list:
            diastereomers.append(2)

        mod_array = []
        for i in range(0, len(diastereomers)):
            mod_array.append(1)
            for j in range(i + 1, len(diastereomers)):
                mod_array[i] *= diastereomers[j]

        prev_diastereomer = 0
        for diastereomer in range(0, int(prod(diastereomers))):
            for i, targ in enumerate(target_list):
                rot = int(diastereomer / mod_array[i]) % diastereomers[i]
                rot -= int(prev_diastereomer / mod_array[i]) % diastereomers[i]
                angle = rot * pi
                if angle != 0:
                    qualifying_fragments = []
                    for atom in targ.connected:
                        frag = geom.get_fragment(atom, targ)
                        if sum([int(targ in frag_atom.connected) for frag_atom in frag]) == 1:
                            qualifying_fragments.append(frag)

                    if len(qualifying_fragments) < 2:
                        qualifying_fragments = []
                        for atom in targ.connected:
                            frag = geom.get_fragment(atom, targ)
                            if sum([int(targ in frag_atom.connected) for frag_atom in frag]) == 2:
                                qualifying_fragments.append(frag)

                        if not qualifying_fragments:
                            warn(
                                "cannot change chirality of atom %s\n" % targ.name +
                                "must have at least two groups not in a ring"
                            )

                        else:
                            frag = qualifying_fragments[0]
                            atom1, atom2 = geom.find(BondedTo(targ), frag)
                            v1 = targ.bond(atom1) / targ.dist(atom1)
                            v2 = targ.bond(atom2) / targ.dist(atom2)
                            rv = v1 + v2
                            geom.rotate(rv, angle=pi, targets=frag, center=targ)

                            if args.minimize:
                                warn(
                                    "cannot minimize steric clashing for cyclic fragment on atom %s" % targ.name
                                )

                    else:
                        while len(qualifying_fragments) > 2:
                            for frag in qualifying_fragments:
                                if len(frag) == max([len(f) for f in qualifying_fragments]):
                                    qualifying_fragments.remove(frag)
                                    break

                        frag1, frag2 = qualifying_fragments
                        atom1 = geom.find(frag1, BondedTo(targ))[0]
                        atom2 = geom.find(frag2, BondedTo(targ))[0]
                        v1 = targ.bond(atom1) / targ.dist(atom1)
                        v2 = targ.bond(atom2) / targ.dist(atom2)
                        rv = v1 + v2

                        geom.rotate(rv, angle=pi, targets=frag1+frag2, center=targ)

                        for frag in [frag1, frag2]:
                            for sub in geom.substituents:
                                if (
                                        all(atom in sub.atoms for atom in frag) and
                                        all(atom in frag for atom in sub.atoms)
                                ):
                                    break
                            else:
                                geom.substituents.append(Substituent(frag, end=targ))

            prev_diastereomer = diastereomer

            if args.minimize:
                geom.minimize_sub_torsion(increment=5)

            if args.outfile is None:
                s += geom.write(outfile=False)
                s += "\n"
            else:
                if os.path.isdir(os.path.expanduser(args.outfile)):
                    outfile = os.path.join(
                        os.path.expanduser(args.outfile),
                        "diastereomer-%i.xyz" % (diastereomer + 1)
                    )

                else:
                    outfile = args.outfile.replace("$i", str(diastereomer + 1))

                geom.write(outfile=outfile, append="$i" not in args.outfile)

    else:
        substituents = []
        for targ in target_list:
            qualifying_fragments = []
            for atom in targ.connected:
                frag = geom.get_fragment(atom, targ)
                if sum([int(targ in frag_atom.connected) for frag_atom in frag]) == 1:
                    qualifying_fragments.append(frag)

            if len(qualifying_fragments) < 2:
                qualifying_fragments = []
                for atom in targ.connected:
                    frag = geom.get_fragment(atom, targ)
                    if sum([int(targ in frag_atom.connected) for frag_atom in frag]) == 2:
                        qualifying_fragments.append(frag)

                if not qualifying_fragments:
                    warn(
                        "cannot change chirality of atom %s\n" % targ.name +
                        "must have at least two groups not in a ring"
                    )

                else:
                    frag = qualifying_fragments[0]
                    atom1, atom2 = geom.find(BondedTo(targ), frag)
                    v1 = targ.bond(atom1) / targ.dist(atom1)
                    v2 = targ.bond(atom2) / targ.dist(atom2)
                    rv = v1 + v2
                    geom.rotate(rv, angle=pi, targets=frag, center=targ)

                    if args.minimize:
                        warn(
                            "cannot minimize steric clashing for cyclic fragment on atom %s" % targ.name
                        )

            else:
                while len(qualifying_fragments) > 2:
                    for frag in qualifying_fragments:
                        if len(frag) == max([len(f) for f in qualifying_fragments]):
                            qualifying_fragments.remove(frag)
                            break

                frag1, frag2 = qualifying_fragments
                atom1 = geom.find(frag1, BondedTo(targ))[0]
                atom2 = geom.find(frag2, BondedTo(targ))[0]
                v1 = targ.bond(atom1) / targ.dist(atom1)
                v2 = targ.bond(atom2) / targ.dist(atom2)
                rv = v1 + v2

                geom.rotate(rv, angle=pi, targets=frag1 + frag2, center=targ)

                if len(frag1) > 1:
                    geom.substituents.append(Substituent(frag1, end=targ))
                if len(frag2) > 1:
                    geom.substituents.append(Substituent(frag2, end=targ))

        if args.minimize:
            geom.minimize_sub_torsion(increment=5)

        if args.outfile is None:
            s += geom.write(outfile=False)
            s += "\n"
        else:
            geom.write(outfile=args.outfile)


if args.outfile is None or args.list_info:
    print(s[:-1])



#! /usr/bin/env python3
from AaronTools.fileIO import FileReader


def main(args):
    fr = FileReader(args.filename, just_geom=False)
    if args.show:
        s = "frequency\t"
        for x in args.show:
            if x == "vector":
                continue
            s += "%s\t" % x
        print(s)
    for i, data in enumerate(sorted(fr.other["frequency"].data, key=lambda x: x.frequency)):
        if args.type == "neg" and data.frequency > 0:
            continue
        if args.type == "pos" and data.frequency < 0:
            continue
        if isinstance(args.type, int) and i + 1 > args.type:
            break
        s = "%9.4f\t" % data.frequency
        for x in args.show:
            if x == "vector":
                continue
            val = getattr(data, x)
            if isinstance(val, float):
                s += "%9.4f\t" % val
            else:
                s += "%s\t" % str(val)
        
        print(s)
        
        if "vector" in args.show:
            print(data.vector)
            print()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Prints frequencies from computational output file",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "filename", help="Completed QM output file with frequency info"
    )
    parser.add_argument(
        "--type",
        "-t",
        type=str,
        nargs="?",
        help="Types of frequencies to print (defaults to `all`. Allowed values: `all`, `neg[ative]`, `pos[itive]`, or `x` to print the first int(x) modes)",
        default="all",
    )
    parser.add_argument(
        "--show",
        "-s",
        type=str,
        nargs="*",
        help="Specify what additional information to show\n"
        "Some info may not be available for certain file formats",
        choices=["intensity", "vector", "forcek", "symmetry"],
        default=[],
    )
    args = parser.parse_args()
    try:
        args.type = int(args.type)
    except ValueError:
        args.type = args.type.lower()
        if args.type not in ["all", "neg", "negative", "pos", "positive"]:
            parser.print_help()
            exit(1)
        elif args.type in ["neg", "negative"]:
            args.type = "neg"
        elif args.type in ["pos", "positive"]:
            args.type = "pos"
    main(args)



#!/usr/bin/env python3

import sys
import argparse
import numpy as np

from AaronTools.fileIO import FileReader, read_types

info_parser = argparse.ArgumentParser(
    description="print information in Gaussian, ORCA, or Psi4 output files",
    formatter_class=argparse.RawTextHelpFormatter
)

info_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

info_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

info_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

info_parser.add_argument(
    "-ls", "--list",
    action="store_true",
    default=False,
    required=False,
    dest="list",
    help="list info categories and exit",
)

info_parser.add_argument(
    "-i", "--info",
    type=str,
    default=[],
    action="append",
    required=False,
    dest="info",
    help="information to print\n" +
    "Default is all info"
)

info_parser.add_argument(
    "-csv",
    "--csv-format",
    nargs="?",
    default=False,
    choices=("comma", "semicolon", "tab", "space"),
    required=False,
    dest="csv",
    help="print info in CSV format with the specified separator\n" +
    "Default: do not print in CSV format",
)

args = info_parser.parse_args()

if args.csv is None:
    args.csv = "comma"

if args.csv:
    if args.csv == "comma":
        sep = ","
    elif args.csv == "tab":
        sep = "\t"
    elif args.csv == "semicolon":
        sep = ";"
    else:
        sep = " "

s = ""

np.set_printoptions(precision=5)

for f in args.infile:
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None), just_geom=False)
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f), just_geom=False)
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f), just_geom=False)

    if args.list:
        s += "info in %s:\n" % f
        for key in infile.other.keys():
            s += "\t%s\n" % key

    else:
        s += "%s:\n" % f
        missing_keys = [key for key in args.info if key not in infile.other.keys()]
        if missing_keys:
            s += "\nmissing some info: %s\n" % ", ".join(missing_keys)

        for key in infile.other.keys():
            if args.info == [] or any(key.lower() == info.lower() for info in args.info):
                if isinstance(infile.other[key], str):
                    if args.csv:
                        s += "\"%s\"%s%s\n" % (key, sep, infile.other[key])
                    else:
                        s += "\t%-30s=\t%s\n" % (key, infile.other[key])
                elif isinstance(infile.other[key], bool):
                    if args.csv:
                        s += "\"%s\"%s%s\n" % (key, sep, infile.other[key])
                    else:
                        s += "\t%-30s =\t%s\n" % (key, str(infile.other[key]))
                elif isinstance(infile.other[key], int):
                    if args.csv:
                        s += "\"%s\"%s%i\n" % (key, sep, infile.other[key])
                    else:
                        s += "\t%-30s =\t%i\n" % (key, infile.other[key])
                elif isinstance(infile.other[key], float):
                    if args.csv:
                        s += "\"%s\"%s%.8f\n" % (key, sep, infile.other[key])
                    else:
                        s += "\t%-30s =\t%.8f\n" % (key, infile.other[key])
                elif isinstance(infile.other[key], list):
                    if args.csv:
                        s += "\"%s\"%s%s\n" % (
                            key, sep, sep.join([str(x) for x in infile.other[key]])
                        )
                    else:
                        s += "\t%-30s =\t%s\n" % (
                            key, ", ".join([str(x) for x in infile.other[key]])
                        )
                elif isinstance(infile.other[key], np.ndarray):
                    if args.csv:
                        s += "\"%s\"%s" % (key, sep)
                        vectorized = np.reshape(infile.other[key], (infile.other[key].size,))
                        if isinstance(vectorized[0], float):
                            s += sep.join(["%11.8f" % x for x in vectorized])
                        else:
                            s += sep.join([str(x) for x in vectorized])
                        s += "\n"
                    else:
                        s += "\t%-30s =\n" % key
                        for line in str(infile.other[key]).splitlines():
                            s += "\t\t%s\n" % line

if not args.outfile:
    print(s.strip())
else:
    with open(args.outfile, "a") as f:
        f.write(s.strip())



"""various job types for Theory() instances"""
import itertools as it

from AaronTools.theory import (
    GAUSSIAN_CONSTRAINTS,
    GAUSSIAN_COORDINATES,
    GAUSSIAN_ROUTE,
    ORCA_BLOCKS,
    ORCA_ROUTE,
    PSI4_BEFORE_GEOM,
    PSI4_COORDINATES,
    PSI4_JOB,
    PSI4_OPTKING,
    PSI4_SETTINGS,
)
from AaronTools.utils.utils import range_list


class JobType:
    """
    parent class of all job types
    initialization keywords should be the same as attribute names
    """

    def __init__(self):
        pass

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.get_psi4() == other.get_psi4()

    def get_gaussian(self):
        """overwrite to return dict with GAUSSIAN_* keys"""
        pass

    def get_orca(self):
        """overwrite to return dict with ORCA_* keys"""
        pass

    def get_psi4(self):
        """overwrite to return dict with PSI4_* keys"""
        pass


class OptimizationJob(JobType):
    """optimization job"""

    def __init__(
        self,
        transition_state=False,
        constraints=None,
        geometry=None,
    ):
        """use transition_state=True to do a TS optimization
        constraints - dict with keys:
                      **** available for ORCA, Gaussian, and Psi4 ****
                      'atoms': atom identifiers/finders - atoms to constrain
                      'bonds': list(atom idenifiers/finders) - distances to constrain
                               each atom identifier in the list should result in exactly 2 atoms
                      'angles': list(atom idenifiers/finders) - 1-3 angles to constrain
                                each atom identifier should result in exactly 3 atoms
                      'torsions': list(atom identifiers/finders) - constrained dihedral angles
                                  each atom identifier should result in exactly 4 atoms
                      **** available for Gaussian and Psi4 ****
                      'x': list(atom identifiers/finders) - constrain the x coordinate of
                           these atoms
                      similarly, 'y' and 'z' are also accepted
                      'xgroup': list(tuple(list(atom idenifiers), x_val, hold)) -
                            constrain the x coordinate of these atoms to be the same
                            x_val - set x-coordinate to this value
                            hold - hold this value constant during the optimization
                                   if 'hold' is omitted, the value will not be held
                                   constant during the optimization
                            e.g. 'xgroup':[("1-6", 0, False), ("13-24", 3.25, False)]
                            this will keep atoms 1-6 and 13-24 in parallel planes, while also
                            allowing those planes to move
                      'ygroup' and 'zgroup' are also available, with analagous options

                      *** NOTE ***
                      for Gaussian, 'bonds', 'angles', and 'torsions' constraints cannot be mixed
                      with 'x', 'y', 'z', 'xgroup', 'ygroup', or 'zgroup' constraints
        geometry    - Geoemtry, will be set when using an AaronTools FileWriter"""
        super().__init__()

        self.transition_state = transition_state
        self.constraints = constraints
        self.geometry = geometry

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE, GAUSSIAN_CONSTRAINTS"""
        if self.transition_state:
            out = {GAUSSIAN_ROUTE: {"Opt": ["ts"]}}
        else:
            out = {GAUSSIAN_ROUTE: {"Opt": []}}

        coords = self.geometry.coords.tolist()
        vars = []
        consts = []
        use_zmat = False

        group_count = 1

        if self.constraints is not None and any(
            self.constraints[key] for key in self.constraints.keys()
        ):
            for key in self.constraints:
                if key not in [
                    "x",
                    "y",
                    "z",
                    "xgroup",
                    "ygroup",
                    "zgroup",
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for Gaussian" % key
                    )
            out[GAUSSIAN_CONSTRAINTS] = []

            if "x" in self.constraints and self.constraints["x"]:
                x_atoms = self.geometry.find(self.constraints["x"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in x_atoms:
                        var_name = "x%i" % (i + 1)
                        consts.append((var_name, atom.coords[0]))
                        coords[i] = [var_name, coords[i][1], coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "y" in self.constraints and self.constraints["y"]:
                y_atoms = self.geometry.find(self.constraints["y"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in y_atoms:
                        var_name = "y%i" % (i + 1)
                        consts.append((var_name, atom.coords[1]))
                        coords[i] = [coords[i][0], var_name, coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "z" in self.constraints and self.constraints["z"]:
                z_atoms = self.geometry.find(self.constraints["z"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in z_atoms:
                        var_name = "z%i" % (i + 1)
                        consts.append((var_name, atom.coords[2]))
                        coords[i] = [coords[i][0], coords[i][1], var_name]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "xgroup" in self.constraints:
                for constraint in self.constraints["xgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    x_atoms = self.geometry.find(finders)
                    var_name = "gx%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in x_atoms:
                            coords[i] = [var_name, coords[i][1], coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "ygroup" in self.constraints:
                for constraint in self.constraints["ygroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    y_atoms = self.geometry.find(finders)
                    var_name = "gy%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in y_atoms:
                            coords[i] = [coords[i][0], var_name, coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "zgroup" in self.constraints:
                for constraint in self.constraints["zgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    z_atoms = self.geometry.find(finders)
                    var_name = "gz%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in z_atoms:
                            coords[i] = [coords[i][0], coords[i][1], var_name]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "atoms" in self.constraints and self.constraints["atoms"]:
                atoms = self.geometry.find(self.constraints["atoms"])
                for atom in atoms:
                    ndx = self.geometry.atoms.index(atom) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append("%2i F" % ndx)
                    else:
                        for j, coord in enumerate(coords[ndx - 1]):
                            if isinstance(coord, str):
                                var_name = coord
                                for k, var in enumerate(vars):
                                    if var[0] == coord and not var[
                                        0
                                    ].startswith("g"):
                                        vars.pop(k)
                                        break
                                else:
                                    var_name = "%s%i" % (
                                        ["x", "y", "z"][j],
                                        ndx,
                                    )
                                    coords[ndx - 1][j] = var_name
                            else:
                                var_name = "%s%i" % (["x", "y", "z"][j], ndx)
                                coords[ndx - 1][j] = var_name
                            if not any(
                                const[0] == var_name for const in consts
                            ):
                                consts.append([var_name, atom.coords[j]])

                if not use_zmat:
                    if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                        out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "bonds" in self.constraints:
                for constraint in self.constraints["bonds"]:
                    atom1, atom2 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "B %2i %2i F" % (ndx1, ndx2)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply bond constraints when using Cartesian Z-Matrix, which"
                            + " is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "angles" in self.constraints:
                for constraint in self.constraints["angles"]:
                    atom1, atom2, atom3 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    ndx3 = self.geometry.atoms.index(atom3) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "A %2i %2i %2i F" % (ndx1, ndx2, ndx3)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply angle constraints when using Cartesian Z-Matrix, which"
                            + " is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "torsions" in self.constraints:
                for constraint in self.constraints["torsions"]:
                    atom1, atom2, atom3, atom4 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    ndx3 = self.geometry.atoms.index(atom3) + 1
                    ndx4 = self.geometry.atoms.index(atom4) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "D %2i %2i %2i %2i F" % (ndx1, ndx2, ndx3, ndx4)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply torsion constraints when using Cartesian Z-Matrix,"
                            + "which is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

        if consts or vars:
            for i, coord in enumerate(coords):
                for j, ax in enumerate(["x", "y", "z"]):
                    if isinstance(coord[j], float):
                        var_name = "%s%i" % (ax, i + 1)
                        vars.append((var_name, coord[j]))
                        coord[j] = var_name

        if consts or vars:
            for coord in coords:
                coord.insert(0, 0)

        out[GAUSSIAN_COORDINATES] = {
            "coords": coords,
            "variables": vars,
            "constants": consts,
        }

        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE, ORCA_BLOCKS"""
        if self.transition_state:
            out = {ORCA_ROUTE: ["OptTS"]}
        else:
            out = {ORCA_ROUTE: ["Opt"]}

        if self.constraints is not None and any(
            self.constraints[key] for key in self.constraints.keys()
        ):
            for key in self.constraints:
                if key not in [
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for ORCA" % key
                    )
            out[ORCA_BLOCKS] = {"geom": ["Constraints"]}
            if "atoms" in self.constraints:
                for constraint in self.constraints["atoms"]:
                    atom1 = self.geometry.find(constraint)[0]
                    ndx1 = self.geometry.atoms.index(atom1)
                    out_str = "    {C %2i C}" % (ndx1)
                    out[ORCA_BLOCKS]["geom"].append(out_str)

            if "bonds" in self.constraints:
                for constraint in self.constraints["bonds"]:
                    atom1, atom2 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    out_str = "    {B %2i %2i C}" % (ndx1, ndx2)
                    out[ORCA_BLOCKS]["geom"].append(out_str)

            if "angles" in self.constraints:
                for constraint in self.constraints["angles"]:
                    atom1, atom2, atom3 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    ndx3 = self.geometry.atoms.index(atom3)
                    out_str = "    {A %2i %2i %2i C}" % (ndx1, ndx2, ndx3)
                    out[ORCA_BLOCKS]["geom"].append(out_str)

            if "torsions" in self.constraints:
                for constraint in self.constraints["torsions"]:
                    atom1, atom2, atom3, atom4 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    ndx3 = self.geometry.atoms.index(atom3)
                    ndx4 = self.geometry.atoms.index(atom4)
                    out_str = "    {D %2i %2i %2i %2i C}" % (
                        ndx1,
                        ndx2,
                        ndx3,
                        ndx4,
                    )
                    out[ORCA_BLOCKS]["geom"].append(out_str)

            out[ORCA_BLOCKS]["geom"].append("end")

        return out

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB, PSI4_OPTKING, PSI4_BEFORE_GEOM"""
        if self.transition_state:
            out = {
                PSI4_JOB: {"optimize": []},
                PSI4_SETTINGS: {"opt_type": ["ts"]},
            }
        else:
            out = {PSI4_JOB: {"optimize": []}}

        coords = self.geometry.coords.tolist()
        vars = []
        group_count = 1

        freeze_str = ""
        freeze_str += 'freeze_list = """\n'
        add_freeze_list = False

        # constraints
        if self.constraints is not None and any(
            [self.constraints[key] for key in self.constraints.keys()]
        ):
            for key in self.constraints:
                if key not in [
                    "x",
                    "y",
                    "z",
                    "xgroup",
                    "ygroup",
                    "zgroup",
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for Psi4" % key
                    )
            out[PSI4_OPTKING] = {}
            if (
                "x" in self.constraints
                and self.constraints["x"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["x"])
                for atom in atoms:
                    freeze_str += "    %2i x\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "y" in self.constraints
                and self.constraints["y"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["y"])
                for atom in atoms:
                    freeze_str += "    %2i y\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "z" in self.constraints
                and self.constraints["z"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["z"])
                for atom in atoms:
                    freeze_str += "    %2i z\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "atoms" in self.constraints
                and self.constraints["atoms"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["atoms"])
                for atom in atoms:
                    freeze_str += "    %2i xyz\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if "xgroup" in self.constraints:
                for constraint in self.constraints["xgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    x_atoms = self.geometry.find(finders)
                    var_name = "gx%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(x_atoms):
                            freeze_str += "    %2i x\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][0] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in x_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if "ygroup" in self.constraints:
                for constraint in self.constraints["ygroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    y_atoms = self.geometry.find(finders)
                    var_name = "gy%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(y_atoms):
                            freeze_str += "    %2i y\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][1] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in y_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if "zgroup" in self.constraints:
                for constraint in self.constraints["zgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    z_atoms = self.geometry.find(finders)
                    var_name = "gz%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(z_atoms):
                            freeze_str += "    %2i z\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][2] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in z_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if add_freeze_list:
                freeze_str += '"""\n'
                freeze_str += "    \n"
                out[PSI4_BEFORE_GEOM] = [freeze_str]
                out[PSI4_OPTKING]["frozen_cartesian"] = ["$freeze_list"]

            if "bonds" in self.constraints:
                if self.constraints["bonds"] and self.geometry is not None:
                    out_str = '("\n'
                    for bond in self.constraints["bonds"]:
                        atom1, atom2 = self.geometry.find(bond)
                        out_str += "        %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_distance"] = [out_str]

            if "angles" in self.constraints:
                if self.constraints["angles"] and self.geometry is not None:
                    out_str = '("\n'
                    for angle in self.constraints["angles"]:
                        atom1, atom2, atom3 = self.geometry.find(angle)
                        out_str += "        %2i %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                            self.geometry.atoms.index(atom3) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_bend"] = [out_str]

            if "torsions" in self.constraints:
                if self.constraints["torsions"] and self.geometry is not None:
                    out_str += '("\n'
                    for torsion in self.constraints["torsions"]:
                        atom1, atom2, atom3, atom4 = self.geometry.find(
                            torsion
                        )
                        out_str += "        %2i %2i %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                            self.geometry.atoms.index(atom3) + 1,
                            self.geometry.atoms.index(atom4) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_dihedral"] = [out_str]

        if vars:
            out[PSI4_COORDINATES] = {
                "coords": coords,
                "variables": vars,
            }

        return out

    def get_xcontrol(self, config, ref=None):
        """
        Generates xcontrol file constraints

        Returns: dict(xcontrol)
        """
        if ref is None:
            ref = "ref.xyz"
        xcontrol = ""
        # only put constraints in xcontrol file so this works with Crest also
        frozen = [i + 1 for i, a in enumerate(self.geometry) if bool(a.flag)]
        if frozen:
            frozen = range_list(frozen)
            xcontrol += "$fix\n"
            xcontrol += "  atoms: {}\n".format(frozen)
            xcontrol += "  freeze: {}\n".format(frozen)
        elif self.constraints:
            xcontrol += "$constrain\n"
            xcontrol += "  force constant={}\n".format(
                config["Job"].get("constrain_force", fallback="0.5")
            )
            xcontrol += "  reference={}\n".format(
                config["Job"].get("constrain_ref", fallback=ref)
            )
            constrained = set([])
            for bond in self.constraints.get("bonds", []):
                bond = self.geometry.find(bond)
                constrained.update(bond)
                xcontrol += "  distance: {},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in bond)
                )
            for angle in self.constraints.get("angles", []):
                angle = self.geometry.find(angle)
                constrained.update(angle)
                xcontrol += "  angle: {},{},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in angle)
                )
            for dihedral in self.constraints.get("torsions", []):
                dihedral = self.geometry.find(dihedral)
                constrained.update(dihedral)
                xcontrol += "  dihedral: {},{},{},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in dihedral)
                )
            relaxed = {
                i + 1
                for i, a in enumerate(self.geometry.atoms)
                if a not in constrained
            }
            relaxed = range_list(relaxed)
            xcontrol += "$metadyn\n"
            xcontrol += "  atoms: {}\n".format(relaxed)
        xcontrol += "$end\n"
        return xcontrol


class FrequencyJob(JobType):
    """frequnecy job"""

    def __init__(self, numerical=False, temperature=None):
        """
        temperature in K for thermochem info, defaults to 298.15 K
        """
        super().__init__()
        if temperature is None:
            temperature = 298.15
        self.numerical = numerical
        self.temperature = temperature

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        out = {
            GAUSSIAN_ROUTE: {
                "Freq": ["temperature=%.2f" % float(self.temperature)]
            }
        }
        if self.numerical:
            out[GAUSSIAN_ROUTE]["Freq"].append("Numerical")

        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        out = {ORCA_BLOCKS: {"freq": ["Temp    %.2f" % self.temperature]}}
        if self.numerical:
            out[ORCA_ROUTE] = ["NumFreq"]
        else:
            out[ORCA_ROUTE] = ["Freq"]

        return out

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        out = {
            PSI4_JOB: {"frequencies": []},
            PSI4_SETTINGS: {"T": ["%.2f" % self.temperature]},
        }
        if self.numerical:
            out[PSI4_JOB]["frequencies"].append('dertype="gradient"')

        return out


class SinglePointJob(JobType):
    """single point energy"""

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        return {GAUSSIAN_ROUTE: {"SP": []}}

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        return {ORCA_ROUTE: ["SP"]}

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        return {PSI4_JOB: {"energy": []}}


class ForceJob(JobType):
    """force/gradient job"""

    def __init__(self, numerical=False):
        super().__init__()
        self.numerical = numerical

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        out = {GAUSSIAN_ROUTE: {"force": []}}
        if self.numerical:
            out[GAUSSIAN_ROUTE]["force"].append("EnGrad")
        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        return {ORCA_ROUTE: ["NumGrad" if self.numerical else "EnGrad"]}

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        out = {PSI4_JOB: {"gradient": []}}
        if self.numerical:
            out[PSI4_JOB]["gradient"].append("dertype='energy'")
        return out



"""used to specify implicit solvent info for Theory()"""
from AaronTools.theory import GAUSSIAN_ROUTE, ORCA_BLOCKS, ORCA_ROUTE


class ImplicitSolvent:
    """implicit solvent info"""

    KNOWN_GAUSSIAN_SOLVENTS = [
        "Water",
        "Acetonitrile",
        "Methanol",
        "Ethanol",
        "IsoQuinoline",
        "Quinoline",
        "Chloroform",
        "DiethylEther",
        "DichloroMethane",
        "DiChloroEthane",
        "CarbonTetraChloride",
        "Benzene",
        "Toluene",
        "ChloroBenzene",
        "NitroMethane",
        "Heptane",
        "CycloHexane",
        "Aniline",
        "Acetone",
        "TetraHydroFuran",
        "DiMethylSulfoxide",
        "Argon",
        "Krypton",
        "Xenon",
        "n-Octanol",
        "1,1,1-TriChloroEthane",
        "1,1,2-TriChloroEthane",
        "1,2,4-TriMethylBenzene",
        "1,2-DiBromoEthane",
        "1,2-EthaneDiol",
        "1,4-Dioxane",
        "1-Bromo-2-MethylPropane",
        "1-BromoOctane",
        "1-BromoPentane",
        "1-BromoPropane",
        "1-Butanol",
        "1-ChloroHexane",
        "1-ChloroPentane",
        "1-ChloroPropane",
        "1-Decanol",
        "1-FluoroOctane",
        "1-Heptanol",
        "1-Hexanol",
        "1-Hexene",
        "1-Hexyne",
        "1-IodoButane",
        "1-IodoHexaDecane",
        "1-IodoPentane",
        "1-IodoPropane",
        "1-NitroPropane",
        "1-Nonanol",
        "1-Pentanol",
        "1-Pentene",
        "1-Propanol",
        "2,2,2-TriFluoroEthanol",
        "2,2,4-TriMethylPentane",
        "2,4-DiMethylPentane",
        "2,4-DiMethylPyridine",
        "2,6-DiMethylPyridine",
        "2-BromoPropane",
        "2-Butanol",
        "2-ChloroButane",
        "2-Heptanone",
        "2-Hexanone",
        "2-MethoxyEthanol",
        "2-Methyl-1-Propanol",
        "2-Methyl-2-Propanol",
        "2-MethylPentane",
        "2-MethylPyridine",
        "2-NitroPropane",
        "2-Octanone",
        "2-Pentanone",
        "2-Propanol",
        "2-Propen-1-ol",
        "3-MethylPyridine",
        "3-Pentanone",
        "4-Heptanone",
        "4-Methyl-2-Pentanone",
        "4-MethylPyridine",
        "5-Nonanone",
        "AceticAcid",
        "AcetoPhenone",
        "a-ChloroToluene",
        "Anisole",
        "Benzaldehyde",
        "BenzoNitrile",
        "BenzylAlcohol",
        "BromoBenzene",
        "BromoEthane",
        "Bromoform",
        "Butanal",
        "ButanoicAcid",
        "Butanone",
        "ButanoNitrile",
        "ButylAmine",
        "ButylEthanoate",
        "CarbonDiSulfide",
        "Cis-1,2-DiMethylCycloHexane",
        "Cis-Decalin",
        "CycloHexanone",
        "CycloPentane",
        "CycloPentanol",
        "CycloPentanone",
        "Decalin-mixture",
        "DiBromoMethane",
        "DiButylEther",
        "DiEthylAmine",
        "DiEthylSulfide",
        "DiIodoMethane",
        "DiIsoPropylEther",
        "DiMethylDiSulfide",
        "DiPhenylEther",
        "DiPropylAmine",
        "E-1,2-DiChloroEthene",
        "E-2-Pentene",
        "EthaneThiol",
        "EthylBenzene",
        "EthylEthanoate",
        "EthylMethanoate",
        "EthylPhenylEther",
        "FluoroBenzene",
        "Formamide",
        "FormicAcid",
        "HexanoicAcid",
        "IodoBenzene",
        "IodoEthane",
        "IodoMethane",
        "IsoPropylBenzene",
        "m-Cresol",
        "Mesitylene",
        "MethylBenzoate",
        "MethylButanoate",
        "MethylCycloHexane",
        "MethylEthanoate",
        "MethylMethanoate",
        "MethylPropanoate",
        "m-Xylene",
        "n-ButylBenzene",
        "n-Decane",
        "n-Dodecane",
        "n-Hexadecane",
        "n-Hexane",
        "NitroBenzene",
        "NitroEthane",
        "n-MethylAniline",
        "n-MethylFormamide-mixture",
        "n,n-DiMethylAcetamide",
        "n,n-DiMethylFormamide",
        "n-Nonane",
        "n-Octane",
        "n-Pentadecane",
        "n-Pentane",
        "n-Undecane",
        "o-ChloroToluene",
        "o-Cresol",
        "o-DiChloroBenzene",
        "o-NitroToluene",
        "o-Xylene",
        "Pentanal",
        "PentanoicAcid",
        "PentylAmine",
        "PentylEthanoate",
        "PerFluoroBenzene",
        "p-IsoPropylToluene",
        "Propanal",
        "PropanoicAcid",
        "PropanoNitrile",
        "PropylAmine",
        "PropylEthanoate",
        "p-Xylene",
        "Pyridine",
        "sec-ButylBenzene",
        "tert-ButylBenzene",
        "TetraChloroEthene",
        "TetraHydroThiophene-s,s-dioxide",
        "Tetralin",
        "Thiophene",
        "Thiophenol",
        "trans-Decalin",
        "TriButylPhosphate",
        "TriChloroEthene",
        "TriEthylAmine",
        "Xylene-mixture",
        "Z-1,2-DiChloroEthene",
    ]

    KNOWN_GAUSSIAN_MODELS = [
        "SMD",
        "CPCM",
        "PCM",
        "DIPOLE",
        "IPCM",
        "ISODENSITY",
        "IEFPCM",
        "SCIPCM",
    ]

    KNOWN_ORCA_CPCM_SOLVENTS = [
        "Water",
        "Acetonitrile",
        "Acetone",
        "Ammonia",
        "Ethanol",
        "Methanol",
        "CH2Cl2",
        "CCl4",
        "DMF",
        "DMSO",
        "Pyridine",
        "THF",
        "Chloroform",
        "Hexane",
        "Benzene",
        "CycloHexane",
        "Octanol",
        "Toluene",
    ]

    KNOWN_ORCA_SMD_SOLVENTS = [
        "1,1,1-TRICHLOROETHANE",
        "CYCLOPENTANE",
        "1,1,2-TRICHLOROETHANE",
        "CYCLOPENTANOL",
        "1,2,4-TRIMETHYLBENZENE",
        "CYCLOPENTANONE",
        "1,2-DIBROMOETHANE",
        "DECALIN (CIS/TRANS MIXTURE)",
        "1,2-DICHLOROETHANE",
        "CIS-DECALIN",
        "1,2-ETHANEDIOL",
        "N-DECANE",
        "1,4-DIOXANE",
        "DIBROMOMETHANE",
        "1-BROMO-2-METHYLPROPANE",
        "DIBUTYLETHER",
        "1-BROMOOCTANE",
        "O-DICHLOROBENZENE",
        "1-BROMOPENTANE",
        "E-1,2-DICHLOROETHENE",
        "1-BROMOPROPANE",
        "Z-1,2-DICHLOROETHENE",
        "1-BUTANOL",
        "DICHLOROMETHANE",
        "1-CHLOROHEXANE",
        "DIETHYL ETHER",
        "1-CHLOROPENTANE",
        "DIETHYL SULFIDE",
        "1-CHLOROPROPANE",
        "DIETHYLAMINE",
        "1-DECANOL",
        "DIIODOMETHANE",
        "1-FLUOROOCTANE",
        "DIISOPROPYL ETHER",
        "1-HEPTANOL",
        "CIS-1,2-DIMETHYLCYCLOHEXANE",
        "1-HEXANOL",
        "DIMETHYL DISULFIDE",
        "1-HEXENE",
        "N,N-DIMETHYLACETAMIDE",
        "1-HEXYNE",
        "N,N-DIMETHYLFORMAMIDE",
        "DMF",
        "1-IODOBUTANE",
        "DIMETHYLSULFOXIDE",
        "DMSO",
        "1-IODOHEXADECANE",
        "DIPHENYLETHER",
        "1-IODOPENTANE",
        "DIPROPYLAMINE",
        "1-IODOPROPANE",
        "N-DODECANE",
        "1-NITROPROPANE",
        "ETHANETHIOL",
        "1-NONANOL",
        "ETHANOL",
        "1-OCTANOL",
        "ETHYL ETHANOATE",
        "1-PENTANOL",
        "ETHYL METHANOATE",
        "1-PENTENE",
        "ETHYL PHENYL ETHER",
        "1-PROPANOL",
        "ETHYLBENZENE",
        "2,2,2-TRIFLUOROETHANOL",
        "FLUOROBENZENE",
        "2,2,4-TRIMETHYLPENTANE",
        "FORMAMIDE",
        "2,4-DIMETHYLPENTANE",
        "FORMIC ACID",
        "2,4-DIMETHYLPYRIDINE",
        "N-HEPTANE",
        "2,6-DIMETHYLPYRIDINE",
        "N-HEXADECANE",
        "2-BROMOPROPANE",
        "N-HEXANE",
        "2-BUTANOL",
        "HEXANOIC ACID",
        "2-CHLOROBUTANE",
        "IODOBENZENE",
        "2-HEPTANONE",
        "IODOETHANE",
        "2-HEXANONE",
        "IODOMETHANE",
        "2-METHOXYETHANOL",
        "ISOPROPYLBENZENE",
        "2-METHYL-1-PROPANOL",
        "P-ISOPROPYLTOLUENE",
        "2-METHYL-2-PROPANOL",
        "MESITYLENE",
        "2-METHYLPENTANE",
        "METHANOL",
        "2-METHYLPYRIDINE",
        "METHYL BENZOATE",
        "2-NITROPROPANE",
        "METHYL BUTANOATE",
        "2-OCTANONE",
        "METHYL ETHANOATE",
        "2-PENTANONE",
        "METHYL METHANOATE",
        "2-PROPANOL",
        "METHYL PROPANOATE",
        "2-PROPEN-1-OL",
        "N-METHYLANILINE",
        "E-2-PENTENE",
        "METHYLCYCLOHEXANE",
        "3-METHYLPYRIDINE",
        "N-METHYLFORMAMIDE (E/Z MIXTURE)",
        "3-PENTANONE",
        "NITROBENZENE",
        "PhNO2",
        "4-HEPTANONE",
        "NITROETHANE",
        "4-METHYL-2-PENTANONE",
        "NITROMETHANE",
        "MeNO2",
        "4-METHYLPYRIDINE",
        "O-NITROTOLUENE",
        "5-NONANONE",
        "N-NONANE",
        "ACETIC ACID",
        "N-OCTANE",
        "ACETONE",
        "N-PENTADECANE",
        "ACETONITRILE",
        "MeCN",
        "PENTANAL",
        "ACETOPHENONE",
        "N-PENTANE",
        "ANILINE",
        "PENTANOIC ACID",
        "ANISOLE",
        "PENTYL ETHANOATE",
        "BENZALDEHYDE",
        "PENTYLAMINE",
        "BENZENE",
        "PERFLUOROBENZENE",
        "BENZONITRILE",
        "PROPANAL",
        "BENZYL ALCOHOL",
        "PROPANOIC ACID",
        "BROMOBENZENE",
        "PROPANONITRILE",
        "BROMOETHANE",
        "PROPYL ETHANOATE",
        "BROMOFORM",
        "PROPYLAMINE",
        "BUTANAL",
        "PYRIDINE",
        "BUTANOIC ACID",
        "TETRACHLOROETHENE",
        "BUTANONE",
        "TETRAHYDROFURAN",
        "THF",
        "BUTANONITRILE",
        "TETRAHYDROTHIOPHENE-S,S-DIOXIDE",
        "BUTYL ETHANOATE",
        "TETRALIN",
        "BUTYLAMINE",
        "THIOPHENE",
        "N-BUTYLBENZENE",
        "THIOPHENOL",
        "SEC-BUTYLBENZENE",
        "TOLUENE",
        "TERT-BUTYLBENZENE",
        "TRANS-DECALIN",
        "CARBON DISULFIDE",
        "TRIBUTYLPHOSPHATE",
        "CARBON TETRACHLORIDE",
        "CCl4",
        "TRICHLOROETHENE",
        "CHLOROBENZENE",
        "TRIETHYLAMINE",
        "CHLOROFORM",
        "N-UNDECANE",
        "A-CHLOROTOLUENE",
        "WATER",
        "O-CHLOROTOLUENE",
        "XYLENE (MIXTURE)",
        "M-CRESOL",
        "M-XYLENE",
        "O-CRESOL",
        "O-XYLENE",
        "CYCLOHEXANE",
        "P-XYLENE",
        "CYCLOHEXANONE",
    ]

    KNOWN_ORCA_MODELS = ["SMD", "CPCM", "C-PCM", "PCM"]

    def __init__(self, solvent_model, solvent):
        self.name = solvent_model
        self.solvent = solvent

    def __repr__(self):
        return "%s(%s)" % (self.name.upper(), self.solvent.lower())

    def __eq__(self, other):
        return repr(self) == repr(other)

    def get_gaussian(self):
        """returns dict() with solvent information for gaussian input files"""
        # need to check if solvent model is available
        warnings = []
        if self.solvent.lower() == "gas":
            # all gas, no solvent
            return (dict(), warnings)

        if not any(
                self.name.upper() == model for model in self.KNOWN_GAUSSIAN_MODELS
        ):
            warnings.append(
                "solvent model is not available in Gaussian: %s\nuse one of: %s"
                % (
                    self.name,
                    " ".join(
                        [
                            "SMD",
                            "CPCM",
                            "PCM",
                            "DIPOLE",
                            "IPCM",
                            "ISODENSITY",
                            "IEFPCM",
                            "SCIPCM",
                        ]
                    ),
                )
            )

        # check some orca solvent keywords and switch to gaussian ones
        solvent = self.solvent
        if solvent.lower() == "chcl2":
            solvent = "DiChloroMethane"
        elif solvent.lower() == "ccl4":
            solvent = "CarbonTetraChloride"
        elif solvent.lower() == "THF":
            solvent = "TetraHydroFuran"
        else:
            if not any(
                    solvent.lower() == gaussian_sol.lower()
                    for gaussian_sol in self.KNOWN_GAUSSIAN_SOLVENTS
            ):
                warnings.append(
                    "solvent is unknown to Gaussian: %s\n" % solvent +
                    "see AaronTools.theory.implicit_solvent.KNOWN_GAUSSIAN_SOLVENTS"
                )

        # route option: scrf(model,solvent=solvent name)
        return (
            {GAUSSIAN_ROUTE: {"scrf": [self.name, "solvent=%s" % solvent]}},
            warnings,
        )

    def get_orca(self):
        """returns dict() with solvent information for orca input files"""
        warnings = []
        if self.solvent.lower() == "gas":
            return (dict(), warnings)

        if not any(
                self.name.upper() == model for model in self.KNOWN_ORCA_MODELS
        ):
            warnings.append(
                "solvent model is not available in ORCA: %s\nuse CPCM or SMD"
                % self.name
            )

        out = {}
        cpcm = True
        # route option: CPCM(solvent name)
        # if using smd, add block %cpcm smd true end
        if self.name.upper() == "SMD":
            cpcm = False
            out[ORCA_BLOCKS] = {"cpcm": ["smd    true"]}

        solvent = self.solvent
        # orca has different solvents for cpcm and smd...
        # check both lists, might be able to switch a gaussian keyword to the correct orca one
        if cpcm:
            if solvent.lower() == "dichloromethane":
                solvent = "CH2Cl2"
            elif solvent.lower() == "carbontetrachloride":
                solvent = "CCl4"
            elif solvent.lower() == "tetrahydrofuran":
                solvent = "THF"
            else:
                if not any(
                        solvent.lower() == orca_sol.lower()
                        for orca_sol in self.KNOWN_ORCA_CPCM_SOLVENTS
                ):
                    warnings.append(
                        "solvent is unknown to ORCA: %s\n" % solvent +
                        "see AaronTools.theory.implicit_solvent.KNOWN_ORCA_CPCM_SOLVENTS"
                    )

        else:
            # TODO: look for gaussian/orca pcm solvent names that need to change
            if not any(
                    solvent.lower() == orca_sol.lower()
                    for orca_sol in self.KNOWN_ORCA_SMD_SOLVENTS
            ):
                warnings.append(
                    "solvent is unknown to ORCA: %s\n" % solvent +
                    "see AaronTools.theory.implicit_solvent.KNOWN_ORCA_SMD_SOLVENTS"
                )

        out[ORCA_ROUTE] = ["CPCM(%s)" % solvent]

        return (out, warnings)

    def get_psi4(self):
        """returns dict() with solvent information for psi4 input files"""

        if self.solvent.lower() == "gas":
            return (dict(), warnings)

        raise NotImplementedError(
            "ImplicitSolvent cannot generate Psi4 solvent settings yet"
        )



#!/usr/bin/env python3
import json
import os
import unittest

import numpy as np
from AaronTools import addlogger
from AaronTools.config import Config
from AaronTools.test import TestWithTimer, prefix


@addlogger
class TestConfig(TestWithTimer):
    LOG = None
    USER_SPECIFIC = ["metadata", "infile", (".*", ".*_dir")]
    config_list = [("blank.ini", None), ("HOH.ini", None)]

    def test_init(self):
        for i, (config_name, config) in enumerate(TestConfig.config_list):
            config = Config(
                os.path.join(prefix, "test_files", config_name),
                quiet=True,
                skip_user_default=True,
            )
            TestConfig.config_list[i] = config_name, config

            test = config.as_dict(skip=self.USER_SPECIFIC)
            ref_name = os.path.join(
                prefix, "ref_files", config_name.replace(".ini", "_init.json")
            )
            # with open(ref_name, "w") as f:
            #     json.dump(test, f, indent=2)

            # need this to make sure python->json stuff is consistent,
            # eg: json doesn't distinguish between tuples and lists
            test = json.loads(json.dumps(test))
            self.LOG.debug(json.dumps(test, indent=2))
            with open(ref_name, "r") as f:
                ref = json.load(f)
            self.maxDiff = None
            self.assertDictEqual(ref, test, msg=config_name)

    def test_parse_changes(self):
        ###
        # substitutions with single point of connections
        ###
        simple = Config(
            os.path.join(prefix, "test_files", "simple_subs.ini"),
            quiet=True,
            skip_user_default=True,
        )
        test = json.loads(json.dumps(simple._changes))
        ref_name = simple.infile.replace(".ini", "_changes.json")
        ref_name = ref_name.replace("test_files", "ref_files")
        # with open(ref_name, "w") as f:
        #     json.dump(simple._changes, f, indent=2)
        with open(ref_name, "r") as f:
            ref = json.load(f)
        self.assertDictEqual(ref, test, msg=simple.infile)

        ###
        # for ring-fusing substitutions (multiple connection points)
        ###
        rings = Config(
            os.path.join(prefix, "test_files", "ring_subs.ini"),
            quiet=True,
            skip_user_default=True,
        )
        test = json.loads(json.dumps(rings._changes))
        ref_name = rings.infile.replace(".ini", "_changes.json")
        ref_name = ref_name.replace("test_files", "ref_files")
        # with open(ref_name, "w") as f:
        #     json.dump(rings._changes, f, indent=2)
        with open(ref_name, "r") as f:
            ref = json.load(f)
        self.assertDictEqual(ref, test, msg=simple.infile)

    def test_for_loop(self):
        for_loop = Config(
            os.path.join(prefix, "test_files", "for_loop.ini"),
            quiet=True,
            skip_user_default=True,
        )
        structure_list = for_loop.get_template()
        for structure, kind in structure_list:
            dihedral = round(
                np.rad2deg(structure.dihedral("3", "1", "2", "6")), 6
            )
            if dihedral < 0:
                dihedral += 360
            self.assertEqual(float(structure.name.split(".")[-1]), dihedral)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestConfig("test_init"))
    suite.addTest(TestConfig("test_parse_changes"))
    suite.addTest(TestConfig("test_for_loop"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())
